@@ -52,6 +52,8 @@
 #include <QVector>
 #include <QXmlStreamReader>
 
+#include <memory>
+
 using namespace Tiled;
 using namespace Tiled::Internal;
 
@@ -138,7 +140,7 @@ private:
 
     QString mError;
     QDir mPath;
-    QScopedPointer<Map> mMap;
+    std::unique_ptr<Map> mMap;
     GidMapper mGidMapper;
     bool mReadingExternalTileset;
 
@@ -270,14 +272,16 @@ Map *MapReaderPrivate::readMap()
     const Map::RenderOrder renderOrder =
             renderOrderFromString(renderOrderString);
 
-    const int nextObjectId =
-            atts.value(QLatin1String("nextobjectid")).toInt();
+    const int nextLayerId = atts.value(QLatin1String("nextlayerid")).toInt();
+    const int nextObjectId = atts.value(QLatin1String("nextobjectid")).toInt();
 
     mMap.reset(new Map(orientation, mapWidth, mapHeight, tileWidth, tileHeight, infinite));
     mMap->setHexSideLength(hexSideLength);
     mMap->setStaggerAxis(staggerAxis);
     mMap->setStaggerIndex(staggerIndex);
     mMap->setRenderOrder(renderOrder);
+    if (nextLayerId)
+        mMap->setNextLayerId(nextLayerId);
     if (nextObjectId)
         mMap->setNextObjectId(nextObjectId);
 
@@ -308,7 +312,7 @@ Map *MapReaderPrivate::readMap()
         }
 
         // Fix up sizes of tile objects. This is for backwards compatibility.
-        LayerIterator iterator(mMap.data());
+        LayerIterator iterator(mMap.get());
         while (Layer *layer = iterator.next()) {
             if (ObjectGroup *objectGroup = layer->asObjectGroup()) {
                 for (MapObject *object : *objectGroup) {
@@ -324,7 +328,7 @@ Map *MapReaderPrivate::readMap()
         }
     }
 
-    return mMap.take();
+    return mMap.release();
 }
 
 SharedTileset MapReaderPrivate::readTileset()
@@ -451,13 +455,12 @@ void MapReaderPrivate::readTilesetTile(Tileset &tileset)
         } else if (xml.name() == QLatin1String("image")) {
             ImageReference imageReference = readImage();
             if (imageReference.hasImage()) {
-                QImage image = imageReference.create();
+                QPixmap image = imageReference.create();
                 if (image.isNull()) {
                     if (imageReference.source.isEmpty())
                         xml.raiseError(tr("Error reading embedded image for tile %1").arg(id));
                 }
-                tileset.setTileImage(tile, QPixmap::fromImage(image),
-                                     imageReference.source);
+                tileset.setTileImage(tile, image, imageReference.source);
             }
         } else if (xml.name() == QLatin1String("objectgroup")) {
             ObjectGroup *objectGroup = readObjectGroup();
@@ -702,11 +705,16 @@ void MapReaderPrivate::readTilesetWangSets(Tileset &tileset)
 static void readLayerAttributes(Layer &layer,
                                 const QXmlStreamAttributes &atts)
 {
+    const QStringRef idRef = atts.value(QLatin1String("id"));
     const QStringRef opacityRef = atts.value(QLatin1String("opacity"));
     const QStringRef visibleRef = atts.value(QLatin1String("visible"));
     const QStringRef lockedRef = atts.value(QLatin1String("locked"));
 
     bool ok;
+    const int id = idRef.toInt(&ok);
+    if (ok)
+        layer.setId(id);
+
     const qreal opacity = opacityRef.toDouble(&ok);
     if (ok)
         layer.setOpacity(opacity);
@@ -783,25 +791,10 @@ void MapReaderPrivate::readTileLayerData(TileLayer &tileLayer)
 
     mMap->setLayerDataFormat(layerDataFormat);
 
-    if (mMap->infinite()) {
-        while (xml.readNext() != QXmlStreamReader::Invalid) {
-            if (xml.isEndElement()) {
-                break;
-            } else if (xml.isStartElement()) {
-                if (xml.name() == QLatin1String("chunk")) {
-                    const QXmlStreamAttributes atts = xml.attributes();
-                    int x = atts.value(QLatin1String("x")).toInt();
-                    int y = atts.value(QLatin1String("y")).toInt();
-                    int width = atts.value(QLatin1String("width")).toInt();
-                    int height = atts.value(QLatin1String("height")).toInt();
-
-                    readTileLayerRect(tileLayer, layerDataFormat, encoding, QRect(x, y, width, height));
-                }
-            }
-        }
-    } else {
-        readTileLayerRect(tileLayer, layerDataFormat, encoding, QRect(0, 0, tileLayer.width(), tileLayer.height()));
-    }
+    readTileLayerRect(tileLayer,
+                      layerDataFormat,
+                      encoding,
+                      QRect(0, 0, tileLayer.width(), tileLayer.height()));
 }
 
 void MapReaderPrivate::readTileLayerRect(TileLayer &tileLayer,
@@ -809,6 +802,9 @@ void MapReaderPrivate::readTileLayerRect(TileLayer &tileLayer,
                                          QStringRef encoding,
                                          QRect bounds)
 {
+    Q_ASSERT(xml.isStartElement() && (xml.name() == QLatin1String("data") ||
+                                      xml.name() == QLatin1String("chunk")));
+
     int x = bounds.x();
     int y = bounds.y();
 
@@ -833,6 +829,16 @@ void MapReaderPrivate::readTileLayerRect(TileLayer &tileLayer,
                 }
 
                 xml.skipCurrentElement();
+            } else if (xml.name() == QLatin1String("chunk")) {
+                const QXmlStreamAttributes atts = xml.attributes();
+                int x = atts.value(QLatin1String("x")).toInt();
+                int y = atts.value(QLatin1String("y")).toInt();
+                int width = atts.value(QLatin1String("width")).toInt();
+                int height = atts.value(QLatin1String("height")).toInt();
+
+                // Recursively call for reading this chunk of data
+                readTileLayerRect(tileLayer, layerDataFormat, encoding,
+                                  QRect(x, y, width, height));
             } else {
                 readUnknownElement();
             }
@@ -1006,7 +1012,7 @@ void MapReaderPrivate::readImageLayerImage(ImageLayer &imageLayer)
 
     QUrl sourceUrl = toUrl(source, mPath);
 
-    imageLayer.loadFromImage(QImage(sourceUrl.toLocalFile()), sourceUrl);
+    imageLayer.loadFromImage(sourceUrl);
 
     xml.skipCurrentElement();
 }
@@ -1148,7 +1154,7 @@ TextData MapReaderPrivate::readObjectText()
     TextData textData;
 
     if (atts.hasAttribute(QLatin1String("fontfamily")))
-        textData.font = QFont(atts.value(QLatin1String("fontfamily")).toString());
+        textData.font.setFamily(atts.value(QLatin1String("fontfamily")).toString());
 
     if (atts.hasAttribute(QLatin1String("pixelsize")))
         textData.font.setPixelSize(atts.value(QLatin1String("pixelsize")).toInt());
