@@ -53,7 +53,7 @@ QVariant MapToVariantConverter::toVariant(const Map &map, const QDir &mapDir)
     QVariantMap mapVariant;
 
     mapVariant[QLatin1String("type")] = QLatin1String("map");
-    mapVariant[QLatin1String("version")] = 1.0;
+    mapVariant[QLatin1String("version")] = 1.2;
     mapVariant[QLatin1String("tiledversion")] = QCoreApplication::applicationVersion();
     mapVariant[QLatin1String("orientation")] = orientationToString(map.orientation());
     mapVariant[QLatin1String("renderorder")] = renderOrderToString(map.renderOrder());
@@ -62,6 +62,7 @@ QVariant MapToVariantConverter::toVariant(const Map &map, const QDir &mapDir)
     mapVariant[QLatin1String("tilewidth")] = map.tileWidth();
     mapVariant[QLatin1String("tileheight")] = map.tileHeight();
     mapVariant[QLatin1String("infinite")] = map.infinite();
+    mapVariant[QLatin1String("nextlayerid")] = map.nextLayerId();
     mapVariant[QLatin1String("nextobjectid")] = map.nextObjectId();
 
     addProperties(mapVariant, map.properties());
@@ -127,21 +128,26 @@ QVariant MapToVariantConverter::toVariant(const Tileset &tileset,
 {
     QVariantMap tilesetVariant;
 
-    if (firstGid > 0)
+    if (firstGid > 0) {
         tilesetVariant[QLatin1String("firstgid")] = firstGid;
 
-    const QString &fileName = tileset.fileName();
-    if (!fileName.isEmpty() && firstGid > 0) {
-        QString source = mMapDir.relativeFilePath(fileName);
-        tilesetVariant[QLatin1String("source")] = source;
+        const QString &fileName = tileset.fileName();
+        if (!fileName.isEmpty()) {
+            QString source = mMapDir.relativeFilePath(fileName);
+            tilesetVariant[QLatin1String("source")] = source;
 
-        // Tileset is external, so no need to write any of the stuff below
-        return tilesetVariant;
+            // Tileset is external, so no need to write any of the stuff below
+            return tilesetVariant;
+        }
+    } else {
+        // Include a 'type' property if we are writing the tileset to its own file
+        tilesetVariant[QLatin1String("type")] = QLatin1String("tileset");
+
+        // Include version in external tilesets
+        tilesetVariant[QLatin1String("version")] = 1.2;
+        tilesetVariant[QLatin1String("tiledversion")] = QCoreApplication::applicationVersion();
     }
 
-    // Include a 'type' property if we are writing the tileset to its own file
-    if (firstGid == 0)
-        tilesetVariant[QLatin1String("type")] = QLatin1String("tileset");
 
     tilesetVariant[QLatin1String("name")] = tileset.name();
     tilesetVariant[QLatin1String("tilewidth")] = tileset.tileWidth();
@@ -190,16 +196,13 @@ QVariant MapToVariantConverter::toVariant(const Tileset &tileset,
 
     // Write the properties, terrain, external image, object group and
     // animation for those tiles that have them.
-    QVariantMap tilePropertiesVariant;
-    QVariantMap tilePropertyTypesVariant;
-    QVariantMap tilesVariant;
+    QVariantList tilesVariant;
     for (const Tile *tile  : tileset.tiles()) {
         const Properties properties = tile->properties();
-        if (!properties.isEmpty()) {
-            tilePropertiesVariant[QString::number(tile->id())] = toVariant(properties);
-            tilePropertyTypesVariant[QString::number(tile->id())] = propertyTypesToVariant(properties);
-        }
         QVariantMap tileVariant;
+
+        addProperties(tileVariant, properties);
+
         if (!tile->type().isEmpty())
             tileVariant[QLatin1String("type")] = tile->type();
         if (tile->terrain() != 0xFFFFFFFF) {
@@ -233,12 +236,10 @@ QVariant MapToVariantConverter::toVariant(const Tileset &tileset,
             tileVariant[QLatin1String("animation")] = frameVariants;
         }
 
-        if (!tileVariant.empty())
-            tilesVariant[QString::number(tile->id())] = tileVariant;
-    }
-    if (!tilePropertiesVariant.empty()) {
-        tilesetVariant[QLatin1String("tileproperties")] = tilePropertiesVariant;
-        tilesetVariant[QLatin1String("tilepropertytypes")] = tilePropertyTypesVariant;
+        if (!tileVariant.empty()) {
+            tileVariant[QLatin1String("id")] = tile->id();
+            tilesVariant << tileVariant;
+        }
     }
     if (!tilesVariant.empty())
         tilesetVariant[QLatin1String("tiles")] = tilesVariant;
@@ -329,32 +330,6 @@ QVariant MapToVariantConverter::toVariant(const WangColor &wangColor) const
     return colorVariant;
 }
 
-QVariant MapToVariantConverter::toVariant(const Properties &properties) const
-{
-    QVariantMap variantMap;
-
-    Properties::const_iterator it = properties.constBegin();
-    Properties::const_iterator it_end = properties.constEnd();
-    for (; it != it_end; ++it) {
-        const QVariant value = toExportValue(it.value(), mMapDir);
-        variantMap[it.key()] = value;
-    }
-
-    return variantMap;
-}
-
-QVariant MapToVariantConverter::propertyTypesToVariant(const Properties &properties) const
-{
-    QVariantMap variantMap;
-
-    Properties::const_iterator it = properties.constBegin();
-    Properties::const_iterator it_end = properties.constEnd();
-    for (; it != it_end; ++it)
-        variantMap[it.key()] = typeToName(it.value().userType());
-
-    return variantMap;
-}
-
 QVariant MapToVariantConverter::toVariant(const QList<Layer *> &layers,
                                           Map::LayerDataFormat format) const
 {
@@ -659,6 +634,9 @@ void MapToVariantConverter::addTileLayerData(QVariantMap &variant,
 void MapToVariantConverter::addLayerAttributes(QVariantMap &layerVariant,
                                                const Layer &layer) const
 {
+    if (layer.id() != 0)
+        layerVariant[QLatin1String("id")] = layer.id();
+
     layerVariant[QLatin1String("name")] = layer.name();
     layerVariant[QLatin1String("x")] = layer.x();
     layerVariant[QLatin1String("y")] = layer.y();
@@ -680,8 +658,7 @@ void MapToVariantConverter::addProperties(QVariantMap &variantMap,
     if (properties.isEmpty())
         return;
 
-    QVariantMap propertiesMap;
-    QVariantMap propertyTypesMap;
+    QVariantList propertiesVariantList;
 
     Properties::const_iterator it = properties.constBegin();
     Properties::const_iterator it_end = properties.constEnd();
@@ -689,10 +666,12 @@ void MapToVariantConverter::addProperties(QVariantMap &variantMap,
         int type = it.value().userType();
         const QVariant value = toExportValue(it.value(), mMapDir);
 
-        propertiesMap[it.key()] = value;
-        propertyTypesMap[it.key()] = typeToName(type);
+        QVariantMap propertyVariantMap;
+        propertyVariantMap[QLatin1String("name")] = it.key();
+        propertyVariantMap[QLatin1String("value")] = value;
+        propertyVariantMap[QLatin1String("type")] = typeToName(type);
+        propertiesVariantList << propertyVariantMap;
     }
 
-    variantMap[QLatin1String("properties")] = propertiesMap;
-    variantMap[QLatin1String("propertytypes")] = propertyTypesMap;
+    variantMap[QLatin1String("properties")] = propertiesVariantList;
 }
