@@ -35,7 +35,7 @@
 #include "tilesetmanager.h"
 #include "wangset.h"
 
-#include <QScopedPointer>
+#include <memory>
 
 namespace Tiled {
 
@@ -73,22 +73,25 @@ Map *VariantToMapConverter::toMap(const QVariant &variant,
     const QString renderOrderString = variantMap[QLatin1String("renderorder")].toString();
     Map::RenderOrder renderOrder = renderOrderFromString(renderOrderString);
 
+    const int nextLayerId = variantMap[QLatin1String("nextlayerid")].toInt();
     const int nextObjectId = variantMap[QLatin1String("nextobjectid")].toInt();
 
-    QScopedPointer<Map> map(new Map(orientation,
-                            variantMap[QLatin1String("width")].toInt(),
-                            variantMap[QLatin1String("height")].toInt(),
-                            variantMap[QLatin1String("tilewidth")].toInt(),
-                            variantMap[QLatin1String("tileheight")].toInt(),
-                            variantMap[QLatin1String("infinite")].toInt()));
+    std::unique_ptr<Map> map(new Map(orientation,
+                                     variantMap[QLatin1String("width")].toInt(),
+                                     variantMap[QLatin1String("height")].toInt(),
+                                     variantMap[QLatin1String("tilewidth")].toInt(),
+                                     variantMap[QLatin1String("tileheight")].toInt(),
+                                     variantMap[QLatin1String("infinite")].toInt()));
     map->setHexSideLength(variantMap[QLatin1String("hexsidelength")].toInt());
     map->setStaggerAxis(staggerAxis);
     map->setStaggerIndex(staggerIndex);
     map->setRenderOrder(renderOrder);
+    if (nextLayerId)
+        map->setNextLayerId(nextLayerId);
     if (nextObjectId)
         map->setNextObjectId(nextObjectId);
 
-    mMap = map.data();
+    mMap = map.get();
     map->setProperties(extractProperties(variantMap));
 
     const QString bgColor = variantMap[QLatin1String("backgroundcolor")].toString();
@@ -120,7 +123,7 @@ Map *VariantToMapConverter::toMap(const QVariant &variant,
             tileset->loadImage();
     }
 
-    return map.take();
+    return map.release();
 }
 
 SharedTileset VariantToMapConverter::toTileset(const QVariant &variant,
@@ -148,11 +151,11 @@ ObjectTemplate *VariantToMapConverter::toObjectTemplate(const QVariant &variant,
 Properties VariantToMapConverter::toProperties(const QVariant &propertiesVariant,
                                                const QVariant &propertyTypesVariant) const
 {
-    const QVariantMap propertiesMap = propertiesVariant.toMap();
-    const QVariantMap propertyTypesMap = propertyTypesVariant.toMap();
-
     Properties properties;
 
+    // read object-based format (1.0)
+    const QVariantMap propertiesMap = propertiesVariant.toMap();
+    const QVariantMap propertyTypesMap = propertyTypesVariant.toMap();
     QVariantMap::const_iterator it = propertiesMap.constBegin();
     QVariantMap::const_iterator it_end = propertiesMap.constEnd();
     for (; it != it_end; ++it) {
@@ -164,6 +167,19 @@ Properties VariantToMapConverter::toProperties(const QVariant &propertiesVariant
         properties[it.key()] = value;
     }
 
+    // read array-based format (1.2)
+    const QVariantList propertiesList = propertiesVariant.toList();
+    for (const QVariant &propertyVariant : propertiesList) {
+        const QVariantMap propertyVariantMap = propertyVariant.toMap();
+        const QString propertyName = propertyVariantMap[QLatin1String("name")].toString();
+        const QString propertyType = propertyVariantMap[QLatin1String("type")].toString();
+        const QVariant propertyValue = propertyVariantMap[QLatin1String("value")];
+        int type = nameToType(propertyType);
+        if (type == QVariant::Invalid)
+            type = QVariant::String;
+        properties[propertyName] = fromExportValue(propertyValue, type, mMapDir);
+    }
+
     return properties;
 }
 
@@ -256,20 +272,9 @@ SharedTileset VariantToMapConverter::toTileset(const QVariant &variant)
         terrain->setProperties(extractProperties(terrainMap));
     }
 
-    // Read tiles (everything except their properties)
-    const QVariantMap tilesVariantMap = variantMap[QLatin1String("tiles")].toMap();
-    QVariantMap::const_iterator it = tilesVariantMap.constBegin();
-    for (; it != tilesVariantMap.end(); ++it) {
+    // Reads tile information (everything except the properties)
+    auto readTile = [&](Tile *tile, const QVariantMap &tileVar) {
         bool ok;
-        const int tileId = it.key().toInt();
-        if (tileId < 0) {
-            mError = tr("Invalid (negative) tile id: %1").arg(tileId);
-            return SharedTileset();
-        }
-
-        Tile *tile = tileset->findOrCreateTile(tileId);
-
-        const QVariantMap tileVar = it.value().toMap();
 
         tile->setType(tileVar[QLatin1String("type")].toString());
 
@@ -286,7 +291,7 @@ SharedTileset VariantToMapConverter::toTileset(const QVariant &variant)
         if (ok)
             tile->setProbability(probability);
 
-        imageVariant = tileVar[QLatin1String("image")];
+        QVariant imageVariant = tileVar[QLatin1String("image")];
         if (!imageVariant.isNull()) {
             const QUrl imagePath = toUrl(imageVariant.toString(), mMapDir);
             tileset->setTileImage(tile, QPixmap(imagePath.toLocalFile()), imagePath);
@@ -324,19 +329,50 @@ SharedTileset VariantToMapConverter::toTileset(const QVariant &variant)
             }
             tile->setFrames(frames);
         }
+    };
+
+    // Read tiles (1.0 format)
+    const QVariant tilesVariant = variantMap[QLatin1String("tiles")];
+    const QVariantMap tilesVariantMap = tilesVariant.toMap();
+    QVariantMap::const_iterator it = tilesVariantMap.constBegin();
+    for (; it != tilesVariantMap.end(); ++it) {
+        const int tileId = it.key().toInt();
+        if (tileId < 0) {
+            mError = tr("Invalid (negative) tile id: %1").arg(tileId);
+            return SharedTileset();
+        }
+
+        Tile *tile = tileset->findOrCreateTile(tileId);
+
+        const QVariantMap tileVar = it.value().toMap();
+        readTile(tile, tileVar);
     }
 
-    // Read tile properties
+    // Read tile properties (1.0 format)
     QVariantMap propertiesVariantMap = variantMap[QLatin1String("tileproperties")].toMap();
     QVariantMap propertyTypesVariantMap = variantMap[QLatin1String("tilepropertytypes")].toMap();
     for (it = propertiesVariantMap.constBegin(); it != propertiesVariantMap.constEnd(); ++it) {
         const int tileId = it.key().toInt();
-        const QVariant propertiesVar = it.value();
+        const QVariant &propertiesVar = it.value();
         const QVariant propertyTypesVar = propertyTypesVariantMap.value(it.key());
         const Properties properties = toProperties(propertiesVar, propertyTypesVar);
         tileset->findOrCreateTile(tileId)->setProperties(properties);
     }
 
+    // Read the tiles saved as a list (1.2 format)
+    const QVariantList tilesVariantList = tilesVariant.toList();
+    for (int i = 0; i < tilesVariantList.count(); ++i) {
+        const QVariantMap tileVar = tilesVariantList[i].toMap();
+        const int tileId  = tileVar[QLatin1String("id")].toInt();
+        if (tileId < 0) {
+            mError = tr("Invalid (negative) tile id: %1").arg(tileId);
+            return SharedTileset();
+        }
+        Tile *tile = tileset->findOrCreateTile(tileId);
+        readTile(tile, tileVar);
+        tile->setProperties(extractProperties(tileVar));
+    }
+
     // Read Wang sets
     const QVariantList wangSetVariants = variantMap[QLatin1String("wangsets")].toList();
     for (const QVariant &wangSetVariant : wangSetVariants) {
@@ -357,7 +393,7 @@ WangSet *VariantToMapConverter::toWangSet(const QVariantMap &variantMap, Tileset
     const QString name = variantMap[QLatin1String("name")].toString();
     const int tile = variantMap[QLatin1String("tile")].toInt();
 
-    QScopedPointer<WangSet> wangSet { new WangSet(tileset, name, tile) };
+    std::unique_ptr<WangSet> wangSet { new WangSet(tileset, name, tile) };
 
     wangSet->setProperties(extractProperties(variantMap));
 
@@ -400,7 +436,7 @@ WangSet *VariantToMapConverter::toWangSet(const QVariantMap &variantMap, Tileset
         wangSet->addWangTile(wangTile);
     }
 
-    return wangSet.take();
+    return wangSet.release();
 }
 
 QSharedPointer<WangColor> VariantToMapConverter::toWangColor(const QVariantMap &variantMap,
@@ -450,6 +486,7 @@ Layer *VariantToMapConverter::toLayer(const QVariant &variant)
         layer = toGroupLayer(variantMap);
 
     if (layer) {
+        layer->setId(variantMap[QLatin1String("id")].toInt());
         layer->setProperties(extractProperties(variantMap));
 
         const QPointF offset(variantMap[QLatin1String("offsetx")].toDouble(),
@@ -469,7 +506,7 @@ TileLayer *VariantToMapConverter::toTileLayer(const QVariantMap &variantMap)
     const int startY = variantMap[QLatin1String("starty")].toInt();
     const QVariant dataVariant = variantMap[QLatin1String("data")];
 
-    typedef QScopedPointer<TileLayer> TileLayerPtr;
+    typedef std::unique_ptr<TileLayer> TileLayerPtr;
     TileLayerPtr tileLayer(new TileLayer(name,
                                          variantMap[QLatin1String("x")].toInt(),
                                          variantMap[QLatin1String("y")].toInt(),
@@ -523,12 +560,12 @@ TileLayer *VariantToMapConverter::toTileLayer(const QVariantMap &variantMap)
         }
     }
 
-    return tileLayer.take();
+    return tileLayer.release();
 }
 
 ObjectGroup *VariantToMapConverter::toObjectGroup(const QVariantMap &variantMap)
 {
-    typedef QScopedPointer<ObjectGroup> ObjectGroupPtr;
+    typedef std::unique_ptr<ObjectGroup> ObjectGroupPtr;
     ObjectGroupPtr objectGroup(new ObjectGroup(variantMap[QLatin1String("name")].toString(),
                                                variantMap[QLatin1String("x")].toInt(),
                                                variantMap[QLatin1String("y")].toInt()));
@@ -554,7 +591,7 @@ ObjectGroup *VariantToMapConverter::toObjectGroup(const QVariantMap &variantMap)
     for (const QVariant &objectVariant : objectVariants)
         objectGroup->addObject(toMapObject(objectVariant.toMap()));
 
-    return objectGroup.take();
+    return objectGroup.release();
 }
 
 MapObject *VariantToMapConverter::toMapObject(const QVariantMap &variantMap)
@@ -652,7 +689,7 @@ MapObject *VariantToMapConverter::toMapObject(const QVariantMap &variantMap)
 
 ImageLayer *VariantToMapConverter::toImageLayer(const QVariantMap &variantMap)
 {
-    typedef QScopedPointer<ImageLayer> ImageLayerPtr;
+    typedef std::unique_ptr<ImageLayer> ImageLayerPtr;
     ImageLayerPtr imageLayer(new ImageLayer(variantMap[QLatin1String("name")].toString(),
                                             variantMap[QLatin1String("x")].toInt(),
                                             variantMap[QLatin1String("y")].toInt()));
@@ -674,7 +711,7 @@ ImageLayer *VariantToMapConverter::toImageLayer(const QVariantMap &variantMap)
         imageLayer->loadFromImage(imageSource);
     }
 
-    return imageLayer.take();
+    return imageLayer.release();
 }
 
 GroupLayer *VariantToMapConverter::toGroupLayer(const QVariantMap &variantMap)
@@ -685,7 +722,7 @@ GroupLayer *VariantToMapConverter::toGroupLayer(const QVariantMap &variantMap)
     const qreal opacity = variantMap[QLatin1String("opacity")].toReal();
     const bool visible = variantMap[QLatin1String("visible")].toBool();
 
-    QScopedPointer<GroupLayer> groupLayer(new GroupLayer(name, x, y));
+    std::unique_ptr<GroupLayer> groupLayer(new GroupLayer(name, x, y));
 
     groupLayer->setOpacity(opacity);
     groupLayer->setVisible(visible);
@@ -699,7 +736,7 @@ GroupLayer *VariantToMapConverter::toGroupLayer(const QVariantMap &variantMap)
         groupLayer->addLayer(layer);
     }
 
-    return groupLayer.take();
+    return groupLayer.release();
 }
 
 QPolygonF VariantToMapConverter::toPolygon(const QVariant &variant) const
@@ -723,7 +760,7 @@ TextData VariantToMapConverter::toTextData(const QVariantMap &variant) const
     const int pixelSize = variant[QLatin1String("pixelsize")].toInt();
 
     if (!family.isEmpty())
-        textData.font = QFont(family);
+        textData.font.setFamily(family);
     if (pixelSize > 0)
         textData.font.setPixelSize(pixelSize);
 
@@ -739,7 +776,7 @@ TextData VariantToMapConverter::toTextData(const QVariantMap &variant) const
     if (!colorString.isEmpty())
         textData.color = QColor(colorString);
 
-    Qt::Alignment alignment = 0;
+    Qt::Alignment alignment;
 
     QString hAlignString = variant[QLatin1String("halign")].toString();
     if (hAlignString == QLatin1String("center"))
