@@ -46,6 +46,8 @@
 #include <QtCore/qmath.h>
 #include <QStyle>
 
+#include "qtcompat_p.h"
+
 namespace Tiled {
 namespace Internal {
 
@@ -93,15 +95,15 @@ MapDocumentActionHandler::MapDocumentActionHandler(QObject *parent)
     mActionGroupLayers = new QAction(this);
     mActionUngroupLayers = new QAction(this);
 
-    mActionDuplicateLayer = new QAction(this);
-    mActionDuplicateLayer->setShortcut(tr("Ctrl+Shift+D"));
-    mActionDuplicateLayer->setIcon(
+    mActionDuplicateLayers = new QAction(this);
+    mActionDuplicateLayers->setShortcut(tr("Ctrl+Shift+D"));
+    mActionDuplicateLayers->setIcon(
             QIcon(QLatin1String(":/images/16x16/stock-duplicate-16.png")));
 
-    mActionMergeLayerDown = new QAction(this);
+    mActionMergeLayersDown = new QAction(this);
 
-    mActionRemoveLayer = new QAction(this);
-    mActionRemoveLayer->setIcon(
+    mActionRemoveLayers = new QAction(this);
+    mActionRemoveLayers->setIcon(
             QIcon(QLatin1String(":/images/16x16/edit-delete.png")));
 
     mActionSelectPreviousLayer = new QAction(this);
@@ -110,14 +112,14 @@ MapDocumentActionHandler::MapDocumentActionHandler(QObject *parent)
     mActionSelectNextLayer = new QAction(this);
     mActionSelectNextLayer->setShortcut(tr("Ctrl+PgUp"));
 
-    mActionMoveLayerUp = new QAction(this);
-    mActionMoveLayerUp->setShortcut(tr("Ctrl+Shift+Up"));
-    mActionMoveLayerUp->setIcon(
+    mActionMoveLayersUp = new QAction(this);
+    mActionMoveLayersUp->setShortcut(tr("Ctrl+Shift+Up"));
+    mActionMoveLayersUp->setIcon(
             QIcon(QLatin1String(":/images/16x16/go-up.png")));
 
-    mActionMoveLayerDown = new QAction(this);
-    mActionMoveLayerDown->setShortcut(tr("Ctrl+Shift+Down"));
-    mActionMoveLayerDown->setIcon(
+    mActionMoveLayersDown = new QAction(this);
+    mActionMoveLayersDown->setShortcut(tr("Ctrl+Shift+Down"));
+    mActionMoveLayersDown->setIcon(
             QIcon(QLatin1String(":/images/16x16/go-down.png")));
 
     mActionToggleOtherLayers = new QAction(this);
@@ -125,6 +127,11 @@ MapDocumentActionHandler::MapDocumentActionHandler(QObject *parent)
     mActionToggleOtherLayers->setIcon(
             QIcon(QLatin1String(":/images/16x16/show_hide_others.png")));
 
+    mActionToggleLockOtherLayers = new QAction(this);
+    mActionToggleLockOtherLayers->setShortcut(tr("Ctrl+Shift+L"));
+    mActionToggleLockOtherLayers->setIcon(
+        QIcon(QLatin1String(":/images/16x16/locked.png")));
+
     mActionLayerProperties = new QAction(this);
     mActionLayerProperties->setIcon(
             QIcon(QLatin1String(":images/16x16/document-properties.png")));
@@ -135,9 +142,9 @@ MapDocumentActionHandler::MapDocumentActionHandler(QObject *parent)
     mActionRemoveObjects = new QAction(this);
     mActionRemoveObjects->setIcon(QIcon(QLatin1String(":/images/16x16/edit-delete.png")));
 
-    Utils::setThemeIcon(mActionRemoveLayer, "edit-delete");
-    Utils::setThemeIcon(mActionMoveLayerUp, "go-up");
-    Utils::setThemeIcon(mActionMoveLayerDown, "go-down");
+    Utils::setThemeIcon(mActionRemoveLayers, "edit-delete");
+    Utils::setThemeIcon(mActionMoveLayersUp, "go-up");
+    Utils::setThemeIcon(mActionMoveLayersDown, "go-down");
     Utils::setThemeIcon(mActionLayerProperties, "document-properties");
     Utils::setThemeIcon(mActionRemoveObjects, "edit-delete");
 
@@ -155,14 +162,15 @@ MapDocumentActionHandler::MapDocumentActionHandler(QObject *parent)
     connect(mActionGroupLayers, &QAction::triggered, this, &MapDocumentActionHandler::groupLayers);
     connect(mActionUngroupLayers, &QAction::triggered, this, &MapDocumentActionHandler::ungroupLayers);
 
-    connect(mActionDuplicateLayer, &QAction::triggered, this, &MapDocumentActionHandler::duplicateLayer);
-    connect(mActionMergeLayerDown, &QAction::triggered, this, &MapDocumentActionHandler::mergeLayerDown);
+    connect(mActionDuplicateLayers, &QAction::triggered, this, &MapDocumentActionHandler::duplicateLayers);
+    connect(mActionMergeLayersDown, &QAction::triggered, this, &MapDocumentActionHandler::mergeLayersDown);
     connect(mActionSelectPreviousLayer, &QAction::triggered, this, &MapDocumentActionHandler::selectPreviousLayer);
     connect(mActionSelectNextLayer, &QAction::triggered, this, &MapDocumentActionHandler::selectNextLayer);
-    connect(mActionRemoveLayer, &QAction::triggered, this, &MapDocumentActionHandler::removeLayer);
-    connect(mActionMoveLayerUp, &QAction::triggered, this, &MapDocumentActionHandler::moveLayerUp);
-    connect(mActionMoveLayerDown, &QAction::triggered, this, &MapDocumentActionHandler::moveLayerDown);
+    connect(mActionRemoveLayers, &QAction::triggered, this, &MapDocumentActionHandler::removeLayers);
+    connect(mActionMoveLayersUp, &QAction::triggered, this, &MapDocumentActionHandler::moveLayersUp);
+    connect(mActionMoveLayersDown, &QAction::triggered, this, &MapDocumentActionHandler::moveLayersDown);
     connect(mActionToggleOtherLayers, &QAction::triggered, this, &MapDocumentActionHandler::toggleOtherLayers);
+    connect(mActionToggleLockOtherLayers, &QAction::triggered, this, &MapDocumentActionHandler::toggleLockOtherLayers);
     connect(mActionLayerProperties, &QAction::triggered, this, &MapDocumentActionHandler::layerProperties);
 
     connect(mActionDuplicateObjects, &QAction::triggered, this, &MapDocumentActionHandler::duplicateObjects);
@@ -191,17 +199,18 @@ void MapDocumentActionHandler::retranslateUi()
     mActionAddGroupLayer->setText(tr("&Group Layer"));
     mActionLayerViaCopy->setText(tr("Layer via Copy"));
     mActionLayerViaCut->setText(tr("Layer via Cut"));
-    mActionGroupLayers->setText(tr("&Group Layer"));
-    mActionUngroupLayers->setText(tr("&Ungroup Layer"));
+    mActionGroupLayers->setText(tr("&Group Layers"));
+    mActionUngroupLayers->setText(tr("&Ungroup Layers"));
 
-    mActionDuplicateLayer->setText(tr("&Duplicate Layer"));
-    mActionMergeLayerDown->setText(tr("&Merge Layer Down"));
-    mActionRemoveLayer->setText(tr("&Remove Layer"));
+    mActionDuplicateLayers->setText(tr("&Duplicate Layers"));
+    mActionMergeLayersDown->setText(tr("&Merge Layer Down"));   // todo: make plural after string-freeze
+    mActionRemoveLayers->setText(tr("&Remove Layers"));
     mActionSelectPreviousLayer->setText(tr("Select Pre&vious Layer"));
     mActionSelectNextLayer->setText(tr("Select &Next Layer"));
-    mActionMoveLayerUp->setText(tr("R&aise Layer"));
-    mActionMoveLayerDown->setText(tr("&Lower Layer"));
+    mActionMoveLayersUp->setText(tr("R&aise Layers"));
+    mActionMoveLayersDown->setText(tr("&Lower Layers"));
     mActionToggleOtherLayers->setText(tr("Show/&Hide all Other Layers"));
+    mActionToggleLockOtherLayers->setText(tr("Lock/&Unlock all Other Layers"));
     mActionLayerProperties->setText(tr("Layer &Properties..."));
 }
 
@@ -223,6 +232,8 @@ void MapDocumentActionHandler::setMapDocument(MapDocument *mapDocument)
                 this, &MapDocumentActionHandler::updateActions);
         connect(mapDocument, &MapDocument::currentLayerChanged,
                 this, &MapDocumentActionHandler::updateActions);
+        connect(mapDocument, &MapDocument::selectedLayersChanged,
+                this, &MapDocumentActionHandler::updateActions);
         connect(mapDocument, &MapDocument::selectedAreaChanged,
                 this, &MapDocumentActionHandler::updateActions);
         connect(mapDocument, &MapDocument::selectedObjectsChanged,
@@ -265,40 +276,44 @@ QMenu *MapDocumentActionHandler::createGroupLayerMenu(QWidget *parent) const
     return groupLayerMenu;
 }
 
+/**
+ * Used to check whether we can cut or delete the current tile selection.
+ */
+static bool isTileSelectionLocked(const MapDocument &mapDocument)
+{
+    if (!mapDocument.selectedArea().isEmpty())
+        for (Layer *layer : mapDocument.selectedLayers())
+            if (layer->isTileLayer() && !layer->isUnlocked())
+                return true;
+
+    return false;
+}
+
 void MapDocumentActionHandler::cut()
 {
     if (!mMapDocument)
         return;
 
-    Layer *currentLayer = mMapDocument->currentLayer();
-    if (!currentLayer || !currentLayer->isUnlocked())
+    if (isTileSelectionLocked(*mMapDocument))
         return;
 
-    TileLayer *tileLayer = dynamic_cast<TileLayer*>(currentLayer);
-    const QRegion &selectedArea = mMapDocument->selectedArea();
-    const QList<MapObject*> selectedObjects = mMapDocument->selectedObjects();
-
-    copy();
+    if (!copy())
+        return;
 
     QUndoStack *stack = mMapDocument->undoStack();
     stack->beginMacro(tr("Cut"));
-
-    if (tileLayer && !selectedArea.isEmpty()) {
-        stack->push(new EraseTiles(mMapDocument, tileLayer, selectedArea));
-    } else if (!selectedObjects.isEmpty()) {
-        for (MapObject *mapObject : selectedObjects)
-            stack->push(new RemoveMapObject(mMapDocument, mapObject));
-    }
-
-    selectNone();
-
+    delete_();
     stack->endMacro();
 }
 
-void MapDocumentActionHandler::copy()
+/**
+ * @returns whether anything was copied.
+ */
+bool MapDocumentActionHandler::copy()
 {
     if (mMapDocument)
-        ClipboardManager::instance()->copySelection(mMapDocument);
+        return ClipboardManager::instance()->copySelection(*mMapDocument);
+    return false;
 }
 
 void MapDocumentActionHandler::delete_()
@@ -306,26 +321,41 @@ void MapDocumentActionHandler::delete_()
     if (!mMapDocument)
         return;
 
-    Layer *currentLayer = mMapDocument->currentLayer();
-    if (!currentLayer || !currentLayer->isUnlocked())
+    if (isTileSelectionLocked(*mMapDocument))
         return;
 
-    TileLayer *tileLayer = dynamic_cast<TileLayer*>(currentLayer);
     const QRegion &selectedArea = mMapDocument->selectedArea();
-    const auto selectedObjects = mMapDocument->selectedObjects();
+
+    QList<QUndoCommand*> commands;
+
+    LayerIterator layerIterator(mMapDocument->map(), Layer::TileLayerType);
+    for (Layer *layer : mMapDocument->selectedLayers()) {
+        if (!layer->isTileLayer())
+            continue;
+
+        auto tileLayer = static_cast<TileLayer*>(layer);
+        const QRegion area = selectedArea.intersected(tileLayer->bounds());
+        if (area.isEmpty())                     // nothing to delete
+            continue;
+
+        // Delete the selected part of the layer
+        commands.append(new EraseTiles(mMapDocument, tileLayer, area));
+    }
+
+    commands.append(new RemoveMapObjects(mMapDocument, mMapDocument->selectedObjects()));
 
     QUndoStack *undoStack = mMapDocument->undoStack();
-    undoStack->beginMacro(tr("Delete"));
 
-    if (tileLayer && !selectedArea.isEmpty()) {
-        undoStack->push(new EraseTiles(mMapDocument, tileLayer, selectedArea));
-    } else if (!selectedObjects.isEmpty()) {
-        for (MapObject *mapObject : selectedObjects)
-            undoStack->push(new RemoveMapObject(mMapDocument, mapObject));
+    if (!commands.isEmpty()) {
+        undoStack->beginMacro(tr("Delete"));
+        for (QUndoCommand *command : commands)
+            undoStack->push(command);
     }
 
     selectNone();
-    undoStack->endMacro();
+
+    if (!commands.isEmpty())
+        undoStack->endMacro();
 }
 
 void MapDocumentActionHandler::selectAll()
@@ -333,24 +363,42 @@ void MapDocumentActionHandler::selectAll()
     if (!mMapDocument)
         return;
 
-    Layer *layer = mMapDocument->currentLayer();
-    if (!layer)
-        return;
+    const bool infinite = mMapDocument->map()->infinite();
 
-    if (TileLayer *tileLayer = layer->asTileLayer()) {
-        QRect all = tileLayer->rect();
-        if (mMapDocument->map()->infinite())
-            all = tileLayer->bounds();
+    QRect all;
+    QList<MapObject*> objects;
 
-        if (mMapDocument->selectedArea() == all)
-            return;
+    for (Layer *layer : mMapDocument->selectedLayers()) {
+        if (!layer->isUnlocked())
+            continue;
 
+        switch (layer->layerType()) {
+        case Layer::TileLayerType: {
+            auto tileLayer = static_cast<TileLayer*>(layer);
+            all |= infinite ? tileLayer->bounds() : tileLayer->rect();
+            break;
+        }
+        case Layer::ObjectGroupType: {
+            if (!layer->isUnlocked())
+                continue;
+
+            auto objectGroup = static_cast<ObjectGroup*>(layer);
+            objects.append(objectGroup->objects());
+            break;
+        }
+        case Layer::ImageLayerType:
+        case Layer::GroupLayerType:
+            break;
+        }
+    }
+
+    if (mMapDocument->selectedArea() != all) {
         QUndoCommand *command = new ChangeSelectedArea(mMapDocument, all);
         mMapDocument->undoStack()->push(command);
-    } else if (ObjectGroup *objectGroup = layer->asObjectGroup()) {
-        if (objectGroup->isUnlocked())
-            mMapDocument->setSelectedObjects(objectGroup->objects());
     }
+
+    if (!objects.isEmpty())
+        mMapDocument->setSelectedObjects(objects);
 }
 
 void MapDocumentActionHandler::selectInverse()
@@ -387,19 +435,13 @@ void MapDocumentActionHandler::selectNone()
     if (!mMapDocument)
         return;
 
-    Layer *layer = mMapDocument->currentLayer();
-    if (!layer)
-        return;
-
-    if (layer->asTileLayer()) {
-        if (mMapDocument->selectedArea().isEmpty())
-            return;
-
+    if (!mMapDocument->selectedArea().isEmpty()) {
         QUndoCommand *command = new ChangeSelectedArea(mMapDocument, QRegion());
         mMapDocument->undoStack()->push(command);
-    } else if (layer->isObjectGroup()) {
-        mMapDocument->setSelectedObjects(QList<MapObject*>());
     }
+
+    if (!mMapDocument->selectedObjects().isEmpty())
+        mMapDocument->setSelectedObjects(QList<MapObject*>());
 }
 
 void MapDocumentActionHandler::copyPosition()
@@ -501,7 +543,7 @@ void MapDocumentActionHandler::layerVia(MapDocumentActionHandler::LayerViaVarian
         newObjectGroup->setDrawOrder(currentObjectGroup->drawOrder());
         newObjectGroup->setColor(currentObjectGroup->color());
 
-        for (MapObject *mapObject : selectedObjects) {
+        for (MapObject *mapObject : qAsConst(selectedObjects)) {
             MapObject *clone = mapObject->clone();
             if (variant == ViaCopy)
                 clone->resetId();
@@ -535,8 +577,7 @@ void MapDocumentActionHandler::layerVia(MapDocumentActionHandler::LayerViaVarian
             break;
         }
         case Layer::ObjectGroupType:
-            for (MapObject *oldObject : selectedObjects)
-                undoStack->push(new RemoveMapObject(mMapDocument, oldObject));
+            undoStack->push(new RemoveMapObjects(mMapDocument, selectedObjects));
             break;
         default:
             Q_ASSERT(false);
@@ -555,61 +596,77 @@ void MapDocumentActionHandler::layerVia(MapDocumentActionHandler::LayerViaVarian
 void MapDocumentActionHandler::groupLayers()
 {
     if (mMapDocument)
-        mMapDocument->groupLayer(mMapDocument->currentLayer());
+        mMapDocument->groupLayers(mMapDocument->selectedLayers());
 }
 
 void MapDocumentActionHandler::ungroupLayers()
 {
     if (mMapDocument)
-        mMapDocument->ungroupLayer(mMapDocument->currentLayer());
+        mMapDocument->ungroupLayers(mMapDocument->selectedLayers());
 }
 
-void MapDocumentActionHandler::duplicateLayer()
+void MapDocumentActionHandler::duplicateLayers()
 {
     if (mMapDocument)
-        mMapDocument->duplicateLayer();
+        mMapDocument->duplicateLayers(mMapDocument->selectedLayers());
 }
 
-void MapDocumentActionHandler::mergeLayerDown()
+void MapDocumentActionHandler::mergeLayersDown()
 {
     if (mMapDocument)
-        mMapDocument->mergeLayerDown();
+        mMapDocument->mergeLayersDown(mMapDocument->selectedLayers());
 }
 
 void MapDocumentActionHandler::selectPreviousLayer()
 {
-    if (mMapDocument)
-        mMapDocument->setCurrentLayer(LayerIterator(mMapDocument->currentLayer()).previous());
+    if (!mMapDocument)
+        return;
+
+    if (Layer *previousLayer = LayerIterator(mMapDocument->currentLayer()).previous()) {
+        mMapDocument->setCurrentLayer(previousLayer);
+        mMapDocument->setSelectedLayers({ previousLayer });
+    }
 }
 
 void MapDocumentActionHandler::selectNextLayer()
 {
-    if (mMapDocument)
-        mMapDocument->setCurrentLayer(LayerIterator(mMapDocument->currentLayer()).next());
+    if (!mMapDocument)
+        return;
+
+    if (Layer *nextLayer = LayerIterator(mMapDocument->currentLayer()).next()) {
+        mMapDocument->setCurrentLayer(nextLayer);
+        mMapDocument->setSelectedLayers({ nextLayer });
+    }
 }
 
-void MapDocumentActionHandler::moveLayerUp()
+void MapDocumentActionHandler::moveLayersUp()
 {
     if (mMapDocument)
-        mMapDocument->moveLayerUp(mMapDocument->currentLayer());
+        mMapDocument->moveLayersUp(mMapDocument->selectedLayers());
 }
 
-void MapDocumentActionHandler::moveLayerDown()
+void MapDocumentActionHandler::moveLayersDown()
 {
     if (mMapDocument)
-        mMapDocument->moveLayerDown(mMapDocument->currentLayer());
+        mMapDocument->moveLayersDown(mMapDocument->selectedLayers());
 }
 
-void MapDocumentActionHandler::removeLayer()
+void MapDocumentActionHandler::removeLayers()
 {
     if (mMapDocument)
-        mMapDocument->removeLayer(mMapDocument->currentLayer());
+        mMapDocument->removeLayers(mMapDocument->selectedLayers());
 }
 
 void MapDocumentActionHandler::toggleOtherLayers()
 {
     if (mMapDocument)
-        mMapDocument->toggleOtherLayers(mMapDocument->currentLayer());
+        mMapDocument->toggleOtherLayers(mMapDocument->selectedLayers());
+}
+
+void MapDocumentActionHandler::toggleLockOtherLayers()
+{
+    if (mMapDocument)
+        mMapDocument->toggleLockOtherLayers(mMapDocument->selectedLayers());
 }
 
 void MapDocumentActionHandler::layerProperties()
@@ -650,25 +707,16 @@ void MapDocumentActionHandler::updateActions()
 {
     Map *map = nullptr;
     Layer *currentLayer = nullptr;
+    QList<Layer *> selectedLayers;
     QRegion selection;
     int selectedObjectsCount = 0;
-    bool canMergeDown = false;
 
     if (mMapDocument) {
         map = mMapDocument->map();
         currentLayer = mMapDocument->currentLayer();
+        selectedLayers = mMapDocument->selectedLayers();
         selection = mMapDocument->selectedArea();
         selectedObjectsCount = mMapDocument->selectedObjects().count();
-
-        if (currentLayer) {
-            int currentLayerIndex = currentLayer->siblingIndex();
-            if (currentLayerIndex > 0) {
-                const auto layers = currentLayer->siblings();
-                Layer *upper = layers.at(currentLayerIndex);
-                Layer *lower = layers.at(currentLayerIndex - 1);
-                canMergeDown = lower->canMergeWith(upper);
-            }
-        }
     }
 
     mActionSelectAll->setEnabled(map);
@@ -700,22 +748,25 @@ void MapDocumentActionHandler::updateActions()
     mActionLayerViaCopy->setEnabled(usableSelection);
     mActionLayerViaCut->setEnabled(usableSelection);
 
-    mActionGroupLayers->setEnabled(currentLayer);
-    mActionUngroupLayers->setEnabled(currentLayer && (currentLayer->isGroupLayer() || currentLayer->parentLayer()));
+    mActionGroupLayers->setEnabled(!selectedLayers.isEmpty());
+    mActionUngroupLayers->setEnabled(std::any_of(selectedLayers.begin(), selectedLayers.end(),
+                                                 [] (Layer *layer) { return layer->isGroupLayer() || layer->parentLayer(); }));
 
     const bool hasPreviousLayer = LayerIterator(currentLayer).previous();
     const bool hasNextLayer = LayerIterator(currentLayer).next();
-    const bool canMoveLayerUp = currentLayer && MoveLayer::canMoveUp(*currentLayer);
-    const bool canMoveLayerDown = currentLayer && MoveLayer::canMoveDown(*currentLayer);
+    const bool canMoveLayersUp = !selectedLayers.isEmpty() && MoveLayer::canMoveUp(selectedLayers);
+    const bool canMoveLayersDown = !selectedLayers.isEmpty() && MoveLayer::canMoveDown(selectedLayers);
 
-    mActionDuplicateLayer->setEnabled(currentLayer);
-    mActionMergeLayerDown->setEnabled(canMergeDown);
+    mActionDuplicateLayers->setEnabled(!selectedLayers.isEmpty());
+    mActionMergeLayersDown->setEnabled(std::any_of(selectedLayers.begin(), selectedLayers.end(),
+                                                   [] (Layer *layer) { return layer->canMergeDown(); }));
     mActionSelectPreviousLayer->setEnabled(hasPreviousLayer);
     mActionSelectNextLayer->setEnabled(hasNextLayer);
-    mActionMoveLayerUp->setEnabled(canMoveLayerUp);
-    mActionMoveLayerDown->setEnabled(canMoveLayerDown);
+    mActionMoveLayersUp->setEnabled(canMoveLayersUp);
+    mActionMoveLayersDown->setEnabled(canMoveLayersDown);
     mActionToggleOtherLayers->setEnabled(currentLayer && (hasNextLayer || hasPreviousLayer));
-    mActionRemoveLayer->setEnabled(currentLayer);
+    mActionToggleLockOtherLayers->setEnabled(currentLayer && (hasNextLayer || hasPreviousLayer));
+    mActionRemoveLayers->setEnabled(!selectedLayers.isEmpty());
     mActionLayerProperties->setEnabled(currentLayer);
 
     mActionDuplicateObjects->setEnabled(selectedObjectsCount > 0);
