@@ -24,7 +24,6 @@
 #include "preferences.h"
 
 #include <QAction>
-#include <QActionGroup>
 #include <QShortcut>
 
 using namespace Tiled;
@@ -78,6 +77,9 @@ void ToolManager::setMapDocument(MapDocument *mapDocument)
  */
 QAction *ToolManager::registerTool(AbstractTool *tool)
 {
+    Q_ASSERT(!tool->mToolManager);
+    tool->mToolManager = this;
+
     tool->setMapDocument(mMapDocument);
 
     QAction *toolAction = new QAction(tool->icon(), tool->name(), this);
@@ -95,8 +97,8 @@ QAction *ToolManager::registerTool(AbstractTool *tool)
 
     mActionGroup->addAction(toolAction);
 
-    connect(tool, SIGNAL(enabledChanged(bool)),
-            this, SLOT(toolEnabledChanged(bool)));
+    connect(tool, &AbstractTool::enabledChanged,
+            this, &ToolManager::toolEnabledChanged);
 
     // Select the first added tool
     if (!mSelectedTool && tool->isEnabled()) {
@@ -109,20 +111,22 @@ QAction *ToolManager::registerTool(AbstractTool *tool)
 
 /**
  * Selects the given tool. It should be previously added using registerTool().
+ *
+ * Returns whether the tool was succesfully selected.
  */
-void ToolManager::selectTool(AbstractTool *tool)
+bool ToolManager::selectTool(AbstractTool *tool)
 {
     if (mSelectedTool == tool)
-        return;
+        return true;
 
     if (tool && !tool->isEnabled()) // Refuse to select disabled tools
-        return;
+        return false;
 
     const auto actions = mActionGroup->actions();
     for (QAction *action : actions) {
         if (action->data().value<AbstractTool*>() == tool) {
             action->trigger();
-            return;
+            return true;
         }
     }
 
@@ -130,6 +134,7 @@ void ToolManager::selectTool(AbstractTool *tool)
     for (QAction *action : actions)
         action->setChecked(false);
     setSelectedTool(nullptr);
+    return tool == nullptr;
 }
 
 void ToolManager::actionTriggered(QAction *action)
@@ -190,20 +195,6 @@ void ToolManager::createShortcuts(QWidget *parent)
     }
 }
 
-void ToolManager::setTile(Tile *tile)
-{
-    mTile = tile;
-    if (mSelectedTool)
-        mSelectedTool->setTile(mTile);
-}
-
-void ToolManager::setObjectTemplate(ObjectTemplate *objectTemplate)
-{
-    mObjectTemplate = objectTemplate;
-    if (mSelectedTool)
-        mSelectedTool->setObjectTemplate(mObjectTemplate);
-}
-
 void ToolManager::toolEnabledChanged(bool enabled)
 {
     AbstractTool *tool = qobject_cast<AbstractTool*>(sender());
@@ -267,8 +258,8 @@ void ToolManager::setSelectedTool(AbstractTool *tool)
         return;
 
     if (mSelectedTool) {
-        disconnect(mSelectedTool, SIGNAL(statusInfoChanged(QString)),
-                   this, SIGNAL(statusInfoChanged(QString)));
+        disconnect(mSelectedTool, &AbstractTool::statusInfoChanged,
+                   this, &ToolManager::statusInfoChanged);
     }
 
     mSelectedTool = tool;
@@ -276,9 +267,7 @@ void ToolManager::setSelectedTool(AbstractTool *tool)
 
     if (mSelectedTool) {
         emit statusInfoChanged(mSelectedTool->statusInfo());
-        connect(mSelectedTool, SIGNAL(statusInfoChanged(QString)),
-                this, SIGNAL(statusInfoChanged(QString)));
-        tool->setTile(mTile);
-        tool->setObjectTemplate(mObjectTemplate);
+        connect(mSelectedTool, &AbstractTool::statusInfoChanged,
+                this, &ToolManager::statusInfoChanged);
     }
 }
